# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for managing attributes
#
# pylint: disable=too-many-arguments

import unittest

from okera import context

class AttributesTest(unittest.TestCase):
    @staticmethod
    def _contains_attribute(namespace, key, attributes):
        for attr in attributes:
            if attr.attribute_namespace == namespace and attr.key == key:
                return True
        return False

    @staticmethod
    def _collect_column_attributes(ds):
        result = {}
        for col in ds.schema.cols:
            if not col.attribute_values:
                continue
            for v in col.attribute_values:
                key = v.database
                if v.table:
                    key += '.' + v.table
                    if v.column:
                        key += '.' + v.column
                if key not in result:
                    result[key] = []
                result[key].append(v)
        return result

    def _verify_attr(self, val, namespace, key, db, tbl, col):
        self.assertTrue(val.attribute.attribute_namespace == namespace)
        self.assertTrue(val.attribute.key == key)
        self.assertTrue(val.database == db)
        self.assertTrue(val.table == tbl)
        self.assertTrue(val.column == col)

    def test_basic(self):
        with context().connect() as conn:
            conn.create_attribute('abac_test', 'v1', True)
            conn.create_attribute('abac_test', 'v2', True)
            attributes = conn.list_attributes('abac_test')
            self.assertTrue(len(attributes) >= 2, msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v2', attributes),
                            msg=str(attributes))
            old_len = len(attributes)
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test' in namespaces)

            # Delete the attribute and make sure it is gone
            self.assertTrue(conn.delete_attribute('abac_test', 'v2'))
            attributes = conn.list_attributes('abac_test')
            self.assertTrue(old_len == len(attributes) + 1)
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertFalse(self._contains_attribute('abac_test', 'v2', attributes),
                             msg=str(attributes))
            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test' in namespaces)

    def test_multiple_namespaces(self):
        with context().connect() as conn:
            conn.create_attribute('abac_test1', 'v', True)
            conn.create_attribute('abac_test2', 'v', True)
            conn.create_attribute('abac_test3', 'v', True)

            namespaces = conn.list_attribute_namespaces()
            self.assertTrue('abac_test1' in namespaces)
            self.assertTrue('abac_test2' in namespaces)
            self.assertTrue('abac_test3' in namespaces)

            attributes = conn.list_attributes('abac_test1')
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertFalse(self._contains_attribute('abac_test2', 'v', attributes))

            # List in all namespaces
            attributes = conn.list_attributes()
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

            attributes = conn.list_attributes(None)
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

            attributes = conn.list_attributes('')
            self.assertTrue(self._contains_attribute('abac_test1', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test2', 'v', attributes))
            self.assertTrue(self._contains_attribute('abac_test3', 'v', attributes))

    def test_create_delete(self):
        with context().connect() as conn:
            conn.delete_attribute('abac_test', 'v1')
            conn.delete_attribute('abac_test', 'v2')

            # Try deleting again, should not exist
            self.assertFalse(conn.delete_attribute('abac_test', 'v1'))
            self.assertFalse(conn.delete_attribute('abac_test', 'v2'))

            # Try creating, should not need to specify if exists
            conn.create_attribute('abac_test', 'v1', False)
            conn.create_attribute('abac_test', 'v2', False)
            attributes = conn.list_attributes('abac_test')
            self.assertTrue(len(attributes) >= 2, msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v1', attributes),
                            msg=str(attributes))
            self.assertTrue(self._contains_attribute('abac_test', 'v2', attributes),
                            msg=str(attributes))

            # Try creating without if exists, should fail
            with self.assertRaises(RuntimeError) as ex_ctx:
                conn.create_attribute('abac_test', 'v1', False)
            self.assertTrue('already exists' in str(ex_ctx.exception))

            with self.assertRaises(RuntimeError) as ex_ctx:
                conn.create_attribute('abaC_test', 'v1', False)
            self.assertTrue('already exists' in str(ex_ctx.exception))

            with self.assertRaises(RuntimeError) as ex_ctx:
                conn.create_attribute('ABAC_TEST', 'V1', False)
            self.assertTrue('already exists' in str(ex_ctx.exception))

            with self.assertRaises(RuntimeError) as ex_ctx:
                conn.create_attribute('abac_test', 'v2', False)
            self.assertTrue('already exists' in str(ex_ctx.exception))

            # Try deleting should exist
            self.assertTrue(conn.delete_attribute('abac_test', 'v1'))
            self.assertTrue(conn.delete_attribute('ABAC_test', 'V2'))

    def test_assign_get_attributes(self):
        db = 'attributes_test_db'
        with context().connect() as conn:
            conn.create_attribute('abac_test', 'V1', True)
            conn.create_attribute('ABAC_test', 'v2', True)

            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)
            self.assertTrue(len(conn.list_datasets(db)) == 0)

            # Create 2 tables and 2 views
            conn.execute_ddl('CREATE TABLE %s.t1(c1 int, c2 int, c3 int)' % db)
            conn.execute_ddl('CREATE TABLE %s.t2(c1 int, c2 int, c3 int)' % db)
            conn.execute_ddl('CREATE VIEW %s.v1 AS SELECT * from %s.t1' % (db, db))
            conn.execute_ddl('CREATE VIEW %s.v2 AS SELECT * from %s.t1' % (db, db))
            self.assertTrue(len(conn.list_datasets(db)) == 4)

            # Get the attributes on t1, should be empty
            ds = conn.list_datasets(db, name='t1')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs_by_col = self._collect_column_attributes(ds)
            self.assertTrue(not attrs_by_col)

            # Assign abac_test.v1 to t1
            conn.assign_attribute('abac_TEST', 'v1', db, 't1')
            # TODO: this assign doesn't fail cleanly. The planner fails the request
            # but when the attributes are received, there are duplicates. Fix this
            # in the planner.
            # conn.assign_attribute('abac_test', 'v1', db, 't1', if_not_exists=True)
            ds = conn.list_datasets(db, name='t1')[0]
            attrs = ds.attribute_values
            self.assertTrue(attrs is not None)
            self.assertTrue(len(attrs) == 1)
            self._verify_attr(attrs[0], 'abac_test', 'v1', db, 't1', None)
            self.assertTrue(not self._collect_column_attributes(ds))

            # Assign abac_test.v2 to t1
            conn.assign_attribute('abac_test', 'v2', db, 't1')
            ds = conn.list_datasets(db, name='t1')[0]
            attrs = ds.attribute_values
            self.assertTrue(attrs is not None)
            self.assertTrue(len(attrs) == 2)
            if attrs[0].attribute.key == 'v1':
                self._verify_attr(attrs[0], 'abac_test', 'v1', db, 't1', None)
                self._verify_attr(attrs[1], 'abac_test', 'v2', db, 't1', None)
            else:
                self._verify_attr(attrs[1], 'abac_test', 'v1', db, 't1', None)
                self._verify_attr(attrs[0], 'abac_test', 'v2', db, 't1', None)

            # Assign abac_test.v1 to v1 and v2.c2
            conn.assign_attribute('abac_test', 'v1', db, 'v1')
            conn.assign_attribute('abac_test', 'v1', db, 'v2', 'c2')
            ds = conn.list_datasets(db, name='v1')[0]
            self._verify_attr(ds.attribute_values[0], 'abac_test', 'v1', db, 'v1', None)

            ds = conn.list_datasets(db, name='v2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self._collect_column_attributes(ds)
            self.assertTrue(('%s.v2.c2' % db) in attrs)
            self.assertTrue(('%s.v2.c1' % db) not in attrs)

            # Try again with upper case db
            ds = conn.list_datasets(db.upper(), name='V2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self._collect_column_attributes(ds)
            self.assertTrue(('%s.v2.c2' % db) in attrs)
            self.assertTrue(('%s.v2.c1' % db) not in attrs)

            # Unassign abac_test.v1 from v2.c2
            conn.unassign_attribute('ABAC_test', 'v1', db, 'v2', 'c2')
            ds = conn.list_datasets(db, name='v2')[0]
            self.assertTrue(ds.attribute_values is None)
            attrs = self._collect_column_attributes(ds)
            self.assertTrue(not attrs)
            # Verify the assignment to v1 is still there
            ds = conn.list_datasets(db, name='v1')[0]
            self._verify_attr(ds.attribute_values[0], 'abac_test', 'v1', db, 'v1', None)

            # Unassign abac_test.v1 from v1
            conn.unassign_attribute('abac_test', 'V1', db, 'v1')
            ds = conn.list_datasets(db, name='v1')[0]
            self.assertTrue(ds.attribute_values is None)

    def test_assign_invalid_attributes(self):
        db = 'attributes_test_db'
        with context().connect() as conn:
            conn.create_attribute('abac_test', 'V1', True)

            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % db)
            conn.execute_ddl('CREATE DATABASE %s' % db)

            conn.assign_attribute('abac_test', 'v1', db)
            with self.assertRaises(Exception) as ex_ctx:
                conn.assign_attribute('abac_test', 'not-there', db)
            self.assertTrue('Cannot assign attributes' in str(ex_ctx.exception),
                            msg=str(ex_ctx.exception))

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for Presto in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use

import unittest
import pytest

import prestodb
import requests
import urllib3

from okera import context

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DEFAULT_PRESTO_HOST = 'localhost'
DEFAULT_PRESTO_PORT = 8049
BAD_PRESTO_PORT = 1234

@pytest.mark.filterwarnings('ignore::urllib3.exceptions.InsecureRequestWarning')
@pytest.mark.filterwarnings('ignore:numpy.dtype size changed')
class PrestoTest(unittest.TestCase):
    def test_basic(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_explicit_presto_host(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_host='127.0.0.1',
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_auto_choose_dialect(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_disabled_auto_choose_dialect(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_disabled_presto_dialect(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050) as conn:
            with pytest.raises(ValueError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_enabled_okera_dialect(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='okera')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_bad_presto_connection(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_host='127.0.0.1',
                         presto_port=BAD_PRESTO_PORT) as conn:
            with pytest.raises(requests.exceptions.ConnectionError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_ctx_namespace(self):
        ctx = context(namespace='okera')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_ctx_bad_namespace(self):
        ctx = context(namespace='okera2')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            with pytest.raises(prestodb.exceptions.PrestoUserError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_presto_conn_namespace(self):
        ctx = context(namespace='okera_bad')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT, namespace='okera') as conn:
            results = conn.scan_as_json('select * from okera_sample.whoami',
                                        dialect='presto')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_presto_conn_bad_namespace(self):
        ctx = context(namespace='okera_bad')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT, namespace='okera2') as conn:
            with pytest.raises(prestodb.exceptions.PrestoUserError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_max_records(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json(
                'select * from okera_sample.users', dialect='presto', max_records=100)
            self.assertEqual(len(results), 100)
            results = conn.scan_as_pandas(
                'select * from okera_sample.users', dialect='presto', max_records=100)
            self.assertEqual(len(results), 100)

    def test_max_records_with_less_records(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('select * from okera_sample.users limit 50',
                                        dialect='presto', max_records=100)
            self.assertEqual(len(results), 50)
            results = conn.scan_as_pandas('select * from okera_sample.users limit 50',
                                          dialect='presto', max_records=100)
            self.assertEqual(len(results), 50)

    def test_json_pandas_equality(self):
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            json_results = conn.scan_as_json('select * from okera_sample.users limit 50',
                                             dialect='presto', max_records=100)
            pandas_results = conn.scan_as_pandas(
                'select * from okera_sample.users limit 50',
                dialect='presto', max_records=100)
            self.assertEqual(json_results, pandas_results.to_dict('records'))

    def test_no_auth(self):
        ctx = context()
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            with pytest.raises(RuntimeError):
                conn.scan_as_json('select * from okera_sample.whoami',
                                  dialect='presto')

    def test_ctx_dialect(self):
        ctx = context(dialect='okera')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_ctx_dialect_override(self):
        ctx = context(dialect='presto')
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            results = conn.scan_as_json('okera_sample.whoami', dialect='okera')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'root')

    def test_complex_types(self):
        datasets = [
            'rs_complex.struct_t',
            'rs_complex.multiple_structs_nested',
            'rs_complex.map_t',
            'rs_complex.struct_nested',
            'rs_complex_parquet.map_struct_t',
            'rs_complex_parquet.map_struct_array_t',
        ]
        ctx = context()
        ctx.enable_token_auth(token_str='root')
        with ctx.connect(host='localhost', port=12050,
                         presto_port=DEFAULT_PRESTO_PORT) as conn:
            for dataset in datasets:
                presto_results = conn.scan_as_json(
                    'select * from %s' % dataset, dialect='presto')
                okera_results = conn.scan_as_json(
                    'select * from %s' % dataset, dialect='okera', strings_as_utf8=True)
                self.assertEqual(okera_results, presto_results)



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests against the crawler

# pylint: disable=broad-except
import time
import unittest

from okera import context
from okera.tests import pycerebro_test_common as common

CRAWLER_STATUS_KEY = 'okera.crawler.status'
CRAWLER_DATA_FILE_PATH_KEY = 'okera.crawler.data_file_path'

class CrawlerIntegrationTest(unittest.TestCase):
    def _wait_for_crawler(self, conn, crawler_db, dataset=None, max_sec=15):
        """ Waits for the crawler to find dataset for this crawler

            If dataset is None, then just wait for the crawler to report COMPLETED.
        """
        sleep_time = 1
        if max_sec > 60:
            sleep_time = 3
        if max_sec > 120:
            sleep_time = 5

        elapsed_sec = 0
        datasets = None
        while elapsed_sec < max_sec:
            time.sleep(sleep_time)
            elapsed_sec += 1
            if dataset:
                datasets = conn.list_dataset_names(crawler_db)
                if (crawler_db + '.' + dataset) in datasets:
                    return datasets
            else:
                params = conn.execute_ddl('DESCRIBE DATABASE %s' % crawler_db)
                for p in params:
                    if p[0] == CRAWLER_STATUS_KEY and p[1] == 'COMPLETED':
                        return conn.list_dataset_names(crawler_db)
        if not datasets:
            datasets = conn.list_dataset_names(crawler_db)
        print("Found datasets: " + str(datasets))
        self.fail('Crawler did not find dataset in time.')

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_basic(self, testType=None):
        crawler = 'integration_test'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/tpch-nation-integration-test/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/cerebrodata-test/' +
                        'tpch-nation-integration-test/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, 'tpch_nation_integration_test')

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_not_slow_5k(self):
        # This table has 5K partitions, make sure it finishes reasonably quickly
        location = 's3://cerebrodata-test/part_flat_5000_data_100/'
        # TODO: make this support ADLS when this dataset exists there
        crawler = 'integration_test'
        crawler_db = '_okera_crawler_' + crawler
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            # If it doesn't find it in time, this will fail with a timeout
            # On a dev EC2 vm, this takes ~3 seconds
            self._wait_for_crawler(conn, crawler_db, 'part_flat_5000_data_100', 10)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_flat(self, testType=None):
        crawler = 'integration_test'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/crawler_flat_structure/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/cerebrodata-test/' +
                        'crawler_flat_structure/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))

            # This should only find one dataset since flat structures is not the default
            self._wait_for_crawler(conn, crawler_db, None)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) == 1)
            conn.execute_ddl('DROP TABLE %s.%s' % (crawler_db, 'crawler_flat_structure'))

            # Configure this crawler to accept flat structures
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, 'okera.crawler.single_file_dataset', 'true'))
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, CRAWLER_STATUS_KEY, ''))

            # Crawler should have been retriggered, wait for it, it should now find
            # both datasets
            self._wait_for_crawler(conn, crawler_db, 'atp_csv')
            self._wait_for_crawler(conn, crawler_db, 'topbabynamesbystate_csv')

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_empty(self, testType=None):
        crawler = 'test_empty'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/empty/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/empty_folder/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            self.fail('Directory was not empty, location : ' + location)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_hidden_file(self, testType=None):
        crawler = 'test_hidden_file'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/hidden_files_only/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/hidden_file/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            self.fail('Directory was not empty or hidden file found, location : '
                      + location)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_bad_path_good_scheme(self, testType=None):
        caught = None
        crawler = 'test_bad_path_good_scheme'
        crawler_db = '_okera_crawler_' + crawler
        bucket = 'thispathdoesnotexistcerebro'
        if testType is None:
            location = ('s3://%s/' % bucket)
        elif testType == 'ADLS':
            location = ('adl://%s.azuredatalakestore.net/'% bucket)
        ctx = context()
        with ctx.connect() as conn:
            try:
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            except Exception as e:
                if "is not accessible" not in str(e) and "does not exist" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if(not caught):
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_null_path(self):
        crawler = 'test_null_path'
        caught = None
        crawler_db = '_okera_crawler' + crawler
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            try:
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, None))
            except Exception as e:
                if "is not accessible" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True

            if not caught:
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_bad_scheme(self):
        caught = None
        crawler = 'test_bad_scheme'
        crawler_db = '_okera_crawler' + crawler
        location = 'foo://bucket/path'
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            try:
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            except Exception as e:
                if "No FileSystem for scheme" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if not caught:
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawl_partitioned(self, testType=None):
        crawler = 'test_crawl_partitioned'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/partitions-test2/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/partitions-test/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)

        if len(contents) != 2:
            self.fail("Did not find 2 datasets at location: " + location)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawl_already_registered(self, testType=None):
        crawler = 'test_crawl_already_registered'
        crawler_db = '_okera_crawler_' + crawler
        if testType is None:
            location = 's3://cerebrodata-test/tpch_nation/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/'
                        'cerebrodata-test/tpch_nation/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            if testType is None:
                conn.execute_ddl("CREATE EXTERNAL TABLE IF NOT EXISTS"
                                 "`default`.`tpch_nation_s3`(s smallint"
                                 ", n string, t smallint, d string) LOCATION '%s'"
                                 % location)
            elif testType == 'ADLS':
                conn.execute_ddl("CREATE EXTERNAL TABLE IF NOT EXISTS"
                                 "`default`.`tpch_nation_adls`(s smallint"
                                 ", n string, t smallint, d string) LOCATION '%s'"
                                 % location)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            print(str(contents))
            self.fail("Found unregistered dataset when should have found none: " +
                      location)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawl_duplicate_paths(self, testType=None):
        crawler1 = 'test_crawl_duplicate_paths_1'
        crawler1_db = '_okera_crawler_' + crawler1
        crawler2 = 'test_crawl_duplicate_paths_2'
        crawler2_db = '_okera_crawler_' + crawler2
        if testType is None:
            location = 's3://cerebrodata-test/alltypes-parquet/'
        elif testType == 'ADLS':
            location = ('adl://okeratestdata.azuredatalakestore.net/'
                        'cerebrodata-test/alltypes-parquet/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler1_db)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler2_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler1, location))
            contents1 = self._wait_for_crawler(conn, crawler1_db, None)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler2, location))
            contents2 = self._wait_for_crawler(conn, crawler2_db, None)

            contents1 = [ds.split(".")[1] for ds in contents1]
            contents2 = [ds.split(".")[1] for ds in contents2]

        self.assertEqual(contents1, contents2)

    @staticmethod
    def __get_columns(dataset):
        cols = []
        partition_cols = []
        for c in dataset.schema.cols:
            if c.hidden:
                continue
            if c.is_partitioning_col:
                partition_cols.append(c.name)
            else:
                cols.append(c.name)
        return cols, partition_cols

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_data_partition_col_dup(self):
        # This dataset has the same column name in the partition path and data file.
        # We want to ignore the column in the data file.
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-test-itay/lake4/partitioned2/'

        ctx = context()
        with ctx.connect() as conn:
            # Default crawler allows duplicates
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, 'partitioned2')
            dataset = conn.list_datasets(crawler_db, name='partitioned2')
            cols, partition_cols = self.__get_columns(dataset[0])
            self.assertTrue('str2' in cols)
            self.assertTrue('str2' in partition_cols)
            show = conn.execute_ddl(
                'show create table %s.%s' % (crawler_db, 'partitioned2'))[0][0]
            self.assertEqual(2, show.count('str2 STRING'), msg=show)

            # Configure this crawler to ignore duplicates
            conn.execute_ddl("DROP TABLE %s.%s" % (crawler_db, 'partitioned2'))
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, 'okera.allow.partition_cols_in_data', 'true'))
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, CRAWLER_STATUS_KEY, ''))
            self._wait_for_crawler(conn, crawler_db, 'partitioned2')
            dataset = conn.list_datasets(crawler_db, name='partitioned2')
            cols, partition_cols = self.__get_columns(dataset[0])

            # str2 should not be in cols now, only in partition cols
            self.assertTrue('str2' not in cols)
            self.assertTrue('str2' in partition_cols)
            show = conn.execute_ddl(
                'show create table %s.%s' % (crawler_db, 'partitioned2'))[0][0]
            self.assertEqual(1, show.count('str2 STRING'), msg=show)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawler_no_suffix(self):
        # Crawl a directory where the datasets don't have file suffixes and make
        # sure we can find them another way.
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-ui/autotagger/no-postfix'

        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            dataset = conn.list_datasets(crawler_db, name='csv')[0]
            self.assertEqual("TEXT", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='parquet')[0]
            self.assertEqual("PARQUET", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='avro')[0]
            self.assertEqual("AVRO", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='json')[0]
            self.assertEqual("JSON", dataset.table_format)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawler_tsv(self):
        # Crawl a directory where the datasets don't have file suffixes and make
        # sure we can find them another way.
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-test-kevin/dea_pain_pills/'

        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            dataset = conn.list_datasets(crawler_db, name='dea_pain_pills')[0]
            self.assertEqual("TEXT", dataset.table_format)
            self.assertEqual(42, len(dataset.schema.cols))

    @unittest.skipIf(common.TEST_LEVEL != "ci", "Only running at CI level.")
    def test_root_crawler(self):
        # Crawl the root test bucket, this can take a while.
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/'

        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            # This takes ~330s on an ec2 dev box
            self._wait_for_crawler(conn, crawler_db, max_sec=500)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 250, msg=str(datasets))

    @unittest.skipIf(common.TEST_LEVEL != "ci", "Only running at CI level.")
    def test_ui_crawler(self):
        # Crawl the ui test bucket, this can take a while.
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-ui/'

        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            # Takes 130 on typical ec2 dev box
            self._wait_for_crawler(conn, crawler_db, max_sec=240)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 80)


    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_das_3489(self):
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/put-test2/'
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=300)
            dataset = conn.list_datasets(crawler_db, name='put_test2')[0]
            self.assertEqual("TEXT", dataset.table_format)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_comments(self):
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/crawler-schema-comments/'
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            # This should have comments from the data file
            whitelist = conn.list_datasets(crawler_db, name='avro_whitelist')[0]
            self.assertTrue('This dataset stores' in whitelist.description,
                            msg=str(whitelist))
            whitelist_col = whitelist.schema.cols[1]
            self.assertTrue('Indicates if the user' in whitelist_col.comment,
                            msg=str(whitelist_col))

            # This should have default comments
            default_comments = conn.list_datasets(crawler_db, name='avro_no_comments')[0]
            self.assertTrue('Discovered by Okera crawler' in default_comments.description,
                            msg=str(default_comments))
            default_comments_col = default_comments.schema.cols[1]
            self.assertTrue(default_comments_col.comment is None)

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_autotagger(self):
        crawler = 'e2e_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-datalake/gdpr-avro/transactions/'
        ctx = context()
        with ctx.connect() as conn:
            # Run it a few times to make sure dropping table and assigning existing
            # tags works.
            for _ in range(3):
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
                self._wait_for_crawler(conn, crawler_db)

                # This should have comments from the data file
                transactions = conn.list_datasets(crawler_db, name='transactions')[0]
                self.assertTrue('This dataset stores' in transactions.description,
                                msg=str(transactions))

                # This column should be tagged
                ccn = transactions.schema.cols[5]
                self.assertTrue(ccn.attribute_values is not None)
                self.assertEqual('credit_card_number',
                                 ccn.attribute_values[0].attribute.key)
                self.assertTrue('credit card number' in ccn.comment,
                                msg=str(ccn))

    @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    def test_crawl_ordering(self):
        crawler = 'test_crawl_ordering'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/partitioned_flat_50000/'
        ctx = context()
        with ctx.connect() as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)

            self.assertEqual(contents, [
                '_okera_crawler_test_crawl_ordering.partitioned_flat_50000'])
            dataset = conn.list_datasets(crawler_db, name='partitioned_flat_50000')[0]
            datafile_path = dataset.metadata[CRAWLER_DATA_FILE_PATH_KEY]
            self.assertEqual(
                datafile_path,
                's3a://cerebrodata-test/partitioned_flat_50000/part=9999/data_0.csv')

    # # TODO: for the following tests we need a way to determine if cluster is configured
    # # with ADLS credentials or not and skip accordingly
    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_basic_adls(self):
    #     self.test_basic(testType='ADLS')

    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_flat_adls(self):
    #     self.test_flat(testType='ADLS')

    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_empty_adls(self):
    #     self.test_empty(testType='ADLS')

    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_hidden_file_adls(self):
    #     self.test_hidden_file(testType='ADLS')

    # # This test takes a long time to execute. See DAS-3203 for details.
    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_bad_path_good_scheme_adls(self):
    #     self.test_bad_path_good_scheme(testType='ADLS')

    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_crawl_partitioned_adls(self):
    #     self.test_crawl_partitioned(testType='ADLS')

    # @unittest.skipIf(common.TEST_LEVEL == "smoke", "Skipping at unit test level")
    # def test_crawl_already_registered_adls(self):
    #     self.test_crawl_already_registered(testType='ADLS')

if __name__ == "__main__":
    unittest.main()



# Copyright 2017 Okera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This packages contains the python compiled bindings from the thrift idl. We consider
# this as part of the client API.
#

# pylint: disable=import-error
# pylint: disable=unused-import
# pylint: disable=wrong-import-position

from __future__ import absolute_import

import os
import sys

from thriftpy import load

# dynamically load the thrift modules
thrift_dir = os.path.join(os.path.dirname(__file__), 'thrift')
OkeraRecordService = load(os.path.join(thrift_dir, 'OkeraRecordService.thrift'),
                            include_dirs=[thrift_dir])
RecordService = load(os.path.join(thrift_dir, 'RecordService.thrift'),
                     include_dirs=[thrift_dir])
sys.modules[OkeraRecordService.__name__] = OkeraRecordService
sys.modules[RecordService.__name__] = RecordService

# import the thrift objects
from RecordService import (
    TGetDatabasesParams, TGetTablesParams, TPlanRequestParams,
    TExecTaskParams, TFetchParams, TNetworkAddress, TRecordFormat, TTypeId,
    TGetPartitionsParams,
    TRecordServiceException, TRequestType, RecordServiceWorker,
    # Internal-only exports (used by rest-server)
    TAttribute,
    TAttributeValue,
    TAttributeValueType,
    TDelegationToken,
    TErrorCode,
    TRecordServiceException)
from OkeraRecordService import (
    OkeraRecordServicePlanner, TExecDDLParams,
    # Internal-only exports (used by rest-server)
    TAccessPermissionLevel,
    TAssignAttributesParams,
    TCreateAttributesParams,
    TDeleteAttributesParams,
    TGetAccessPermissionsParams,
    TGetAttributeNamespacesParams,
    TGetAttributesParams,
    TGetDatasetsParams,
    TGetRegisteredObjectsParams,
    TGetRoleProvenanceParams,
    TGetGrantableRolesParams,
    TListFilesOp,
    TListFilesParams,
    TSetAttributesParams,
    TUnassignAttributesParams)



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=no-else-return


import datetime
import json
import warnings

import pytest
import numpy

import requests

from okera import context

API_URL = "http://localhost:5000"

def get_scan_as_json(conn, dataset):
    data = conn.scan_as_json(
        dataset, strings_as_utf8=True,
        max_records=10, max_client_process_count=1)
    return data

def get_scan_as_pandas(conn, dataset):
    def convert_types(datum):
        if isinstance(datum, datetime.datetime):
            return datum.replace(tzinfo=None).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        elif isinstance(datum, numpy.int64):
            return int(datum)
        elif isinstance(datum, dict):
            for key, value in datum.items():
                datum[key] = convert_types(value)
            return datum
        elif isinstance(datum, list):
            return [convert_types(child) for child in datum]

        return datum

    data = conn.scan_as_pandas(
        dataset, strings_as_utf8=True,
        max_records=10, max_client_process_count=1)
    data = data.replace({numpy.nan: None})
    data = data.to_dict('records')
    return convert_types(data)

def get_scan_from_rest(dataset):
    headers = {'content-type': 'application/json'}
    query = {'query': 'SELECT * FROM %s' % dataset}
    response = requests.post(
        API_URL + '/api/scan?records=10',
        json=query, headers=headers)
    result = json.loads(response.text)
    return result

def format(data):
    return json.dumps(data, sort_keys=True, indent=1)

@pytest.mark.parametrize("dataset", [
    'rs_complex.array_struct_array',
    'rs_complex.array_struct_t',
    'rs_complex.array_t',
    'rs_complex.avrotbl',
    'rs_complex.bytes_type',
    'rs_complex.bytes_type_file',
    'rs_complex.enum_type',
    'rs_complex.enum_type_default',
    'rs_complex.map_t',
    'rs_complex.market_decide_single_avro',
    'rs_complex.market_v20_single',
    'rs_complex.market_v30_single',
    'rs_complex.multiple_structs_nested',
    'rs_complex.strarray_t',
    'rs_complex.strarray_t_view',
    'rs_complex.struct_array_struct',
    'rs_complex.struct_nested',
    'rs_complex.struct_nested_s1',
    'rs_complex.struct_nested_s1_f1',
    'rs_complex.struct_nested_s1_s2',
    'rs_complex.struct_nested_view',
    'rs_complex.struct_t',
    'rs_complex.struct_t2',
    'rs_complex.struct_t3',
    'rs_complex.struct_t_id',
    'rs_complex.struct_t_s1',
    'rs_complex.struct_t_view',
    'rs_complex.struct_t_view2',
    'rs_complex.user_phone_numbers',
    'rs_complex.user_phone_numbers_map',
    'rs_complex.users',
    'rs_complex.view_over_multiple_structs',
    'rs_complex_parquet.array_struct_array',
    'rs_complex_parquet.array_struct_t',
    'rs_complex_parquet.array_struct_t2',
    'rs_complex_parquet.array_t',
    'rs_complex_parquet.bytes_type',
    'rs_complex_parquet.enum_type',
    'rs_complex_parquet.map_array',
    'rs_complex_parquet.map_array_t2',
    'rs_complex_parquet.map_struct_array_t',
    'rs_complex_parquet.map_struct_array_t_view',
    'rs_complex_parquet.map_struct_t',
    'rs_complex_parquet.map_struct_t_view',
    'rs_complex_parquet.map_t',
    'rs_complex_parquet.strarray_t',
    'rs_complex_parquet.struct_array_struct',
    'rs_complex_parquet.struct_nested',
    'rs_complex_parquet.struct_nested_s1',
    'rs_complex_parquet.struct_nested_s1_f1',
    'rs_complex_parquet.struct_nested_s1_s2',
    'rs_complex_parquet.struct_nested_view',
    'rs_complex_parquet.struct_t',
    'rs_complex_parquet.struct_t2',
    'rs_complex_parquet.struct_t3',
    'rs_complex_parquet.struct_t_id',
    'rs_complex_parquet.struct_t_s1',
    'rs_complex_parquet.struct_t_view',
    'rs_complex_parquet.struct_t_view2',
    # # The ones below don't work because the /scan API returns the wrong value
    # # for decimals
    # 'rs_complex_parquet.spark_all_mixed_compression',
    # 'rs_complex_parquet.spark_gzip',
    # 'rs_complex_parquet.spark_snappy',
    # 'rs_complex_parquet.spark_snappy_part',
    # 'rs_complex_parquet.spark_uncompressed',
    # 'rs_complex_parquet.spark_uncompressed_legacy_format',
    'customer.zd277_complex',
    'rs_json_format.complex_c1_case_sensitive',
    'rs_json_format.complex_c1_usecase',
    'rs_json_format.complex_nike_usecase',
    # NOT WORKING array of arrays is not allowed, scan error
    # 'rs_json_format.json_array_arrays',
    'rs_json_format.json_array_struct',
    'rs_json_format.json_arrays_test',
    # # These tables don't work because Pandas and JSON don't serialize to the
    # # exact same types (e.g. float vs int), even though the overall data is correct)
    # 'rs_json_format.json_inferred',
    # 'rs_json_format.json_primitives',
    # 'rs_json_format.json_primitives_inferred',
    'rs_json_format.json_struct',
    'rs_json_format.json_struct_array',
    'rs_json_format.json_struct_nested',
    'rs_json_format.primitives_with_array',
])
def test_basic(dataset):
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    ctx = context()
    with ctx.connect(host='localhost', port=12050) as conn:
        json_data = get_scan_as_json(conn, dataset)
        pandas_data = get_scan_as_pandas(conn, dataset)
        rest_data = get_scan_from_rest(dataset)
        assert format(json_data) == format(pandas_data)
        assert format(json_data) == format(rest_data)



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for auth in PyOkera
#
# pylint: disable=global-statement
# pylint: disable=no-self-use

import unittest
import pytest

import thriftpy
from okera import context

#
# NOTE: this test suite uses the fact that enable_token_auth will treat
# a string that has a period in it as a JWT/OkeraToken, and any other string
# as if it was the username (for an unauthed server). Since the test runs
# against an unauthed server, a JWT/OkeraToken will always fail, causing
# us to go through the retry logic.
#

# Global variable for attempts
attempts = 0

def bad_then_good_func():
    global attempts
    token = None
    if attempts == 0:
        token = "foo.bar"
    else:
        token = "foo"
    attempts += 1
    return token

def bad_twice_then_good_func():
    global attempts
    token = None
    if attempts < 2:
        token = "foo.bar"
    else:
        # should never be called
        raise Exception()
    attempts += 1
    return token

def good_then_bad_func():
    global attempts
    token = None
    if attempts == 0:
        token = "foo"
    else:
        # this should never be called
        raise Exception()
    attempts += 1
    return token

def always_bad():
    return "foo.bar"

class AuthTest(unittest.TestCase):
    @staticmethod
    def _reset_attempts():
        global attempts
        attempts = 0

    def setUp(self):
        AuthTest._reset_attempts()

    def test_token_func_basic(self):
        ctx = context()
        ctx.enable_token_auth(token_func=bad_then_good_func)
        with ctx.connect(host='localhost', port=12050) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_no_failure(self):
        ctx = context()
        ctx.enable_token_auth(token_func=good_then_bad_func)
        with ctx.connect(host='localhost', port=12050) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_and_token_str(self):
        ctx = context()
        ctx.enable_token_auth(token_func=always_bad, token_str='foo')
        with ctx.connect(host='localhost', port=12050) as conn:
            results = conn.scan_as_json('okera_sample.whoami')
            self.assertEqual(len(results), 1)
            self.assertEqual(results[0]['user'], 'foo')

    def test_token_func_never_succeed(self):
        ctx = context()
        ctx.enable_token_auth(token_func=always_bad)
        with pytest.raises(thriftpy.transport.TTransportException):
            with ctx.connect(host='localhost', port=12050):
                # we should never get here
                raise Exception()

    def test_token_func_retry_once_still_fail(self):
        ctx = context()
        ctx.enable_token_auth(token_func=bad_twice_then_good_func)
        with pytest.raises(thriftpy.transport.TTransportException):
            with ctx.connect(host='localhost', port=12050):
                # we should never get here
                raise Exception()

    def test_token_func_non_picklable(self):
        def fn():
            return "foo"

        ctx = context()
        with pytest.raises(ValueError):
            ctx.enable_token_auth(token_func=fn)



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for role granting
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-public-methods

import unittest

from okera import context, _thrift_api

TEST_USER = 'roles_test_user'

def disable_auth(ctx, params):
    ctx.disable_auth()
    params.requesting_user = None

def enable_auth(ctx, params, user):
    ctx.enable_token_auth(token_str=user)
    params.requesting_user = user

def get_grantable_roles(conn, params):
    #pylint: disable=protected-access
    return conn._underlying_client().GetGrantableRoles(params)
    #pylint: enable=protected-access

class RolesTest(unittest.TestCase):
    # NOTE: this test will likely need to change once we can properly
    # ACL roles themselves, since then we will return just a subset of
    # roles.
    def test_get_grantable_roles(self):
        ctx = context()
        with ctx.connect() as conn:
            params = _thrift_api.TGetGrantableRolesParams()

            # Set up a role and grant it to the user's group
            disable_auth(ctx, params)
            conn.execute_ddl("DROP ROLE IF EXISTS grantable_role")
            conn.execute_ddl("CREATE ROLE grantable_role")
            conn.execute_ddl("GRANT ROLE grantable_role to GROUP %s" % TEST_USER)

            # As an admin, we should always be able to see all roles
            disable_auth(ctx, params)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertTrue("grantable_role" in retrieved_roles)

            # Even though we have the test role, our user does not have
            # any roles that have a privilege with GRANT OPTION, so we
            # should not get back any roles
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # As an admin, grant a privilege with GRANT OPTION to the
            # role we created
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON TABLE okera_sample.sample
            TO ROLE grantable_role WITH GRANT OPTION""")
            admin_roles = get_grantable_roles(conn, params).roles

            # Even though there is a grant with GRANT OPTION, it's on
            # a specific table, and we specify no parameters, which is
            # equivalent to searching for grants on the CATALOG
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Now, we will check with specific parameters
            params.database = "okera_sample"
            params.table = "sample"

            # As a user, we should now get back all the roles since we have a
            # privilege that has a GRANT OPTION that matches our scope. The set of
            # roles should be the same as what the admin got.
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Now, we will check with specific parameters on a table we don't have
            params.database = "okera_sample"
            params.table = "users"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Now, we will check with specific parameters on just the DB
            params.database = "okera_sample"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # As an admin, grant a privilege with GRANT OPTION to the
            # role we created
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON DATABASE okera_sample
            TO ROLE grantable_role WITH GRANT OPTION""")

            # As a user, we should now get all roles, as we have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Now, we will check with specific parameters on just the DB,
            # with a different DB
            params.database = "okera_system"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Finally, let's add a grant to the CATALOG and then
            # a query with no parameters will work
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON SERVER
            TO ROLE grantable_role WITH GRANT OPTION""")

            # Now, we will check with no parameters
            params.database = ""
            params.table = ""

            # As a user, we should get all the roles, we have GRANT OPTION
            # at the CATALOG level
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Additionally, even if we query on a specific DB (or even table),
            # we will now always get roles because our CATALOG level grant
            # allows us to see them.
            params.database = "okera_system"

            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for ABAC
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods

import unittest

from okera import context, _thrift_api

TEST_DB = 'abac_test_db'
TEST_DB2 = 'abac_test_db2'
TEST_TBL = 'tbl'
TEST_TBL2 = 'tbl2'
TEST_DB_DROP = 'drop_db1'
TEST_TBL_DROP = 'drop_table1'
TEST_VIEW = 'v'
TEST_ROLE = 'abac_test_role'
TEST_USER = 'abac_test_user'
TEST_ROLE_DROP = 'abac_test_role_drop'
TEST_USER_DROP = 'abac_test_user_drop'

class AbacTest(unittest.TestCase):
    @staticmethod
    def __grant_abac_db(conn, db, namespace, key):
        conn.execute_ddl(
            'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
            (db, namespace, key, TEST_ROLE))

    @staticmethod
    def __grant_abac_tbl(conn, db, tbl, namespace, key):
        conn.execute_ddl(
            'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
            (db, tbl, namespace, key, TEST_ROLE))

    @staticmethod
    def __revoke_abac_db(conn, db, namespace, key):
        conn.execute_ddl(
            'REVOKE SELECT ON DATABASE %s HAVING ATTRIBUTE(%s.%s) FROM ROLE %s' %\
            (db, namespace, key, TEST_ROLE))

    @staticmethod
    def __revoke_abac_tbl(conn, db, tbl, namespace, key):
        conn.execute_ddl(
            'REVOKE SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s.%s) FROM ROLE %s' %\
            (db, tbl, namespace, key, TEST_ROLE))

    @staticmethod
    def __show_grant_abac_role(conn):
        return conn.execute_ddl('SHOW GRANT ROLE %s' % (TEST_ROLE))

    @staticmethod
    def __ddl_count(conn, sql):
        return len(conn.execute_ddl(sql))

    @staticmethod
    def __visible_cols(cols):
        result = []
        for c in cols:
            if c.hidden:
                continue
            result.append(c)
        return result

    def __verify_tbl_access(self, conn, db, tbl, num_cols, has_db_access=False,
                            skip_metadata_check=False):
        """ Verifies the current connect has access to num_cols on this table

            FIXME(BUG): skip_metadata_check should be removed (and always True). It is
            skipping due to existing bugs.
        """
        if num_cols == 0:
            for ddl in ['describe %s.%s', 'describe formatted %s.%s']:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    print(conn.execute_ddl(ddl % (db, tbl)))
                self.assertTrue('does not have privilege' in str(ex_ctx.exception))

            if has_db_access:
                # User has access to DB, make sure table does not show up in list
                self.assertTrue('%s.%s' %(db, tbl) not in conn.list_dataset_names(db))
                names = conn.list_dataset_names(db)
                self.assertFalse('%s.%s' %(db, tbl) in names, msg=str(names))
            else:
                # Doesn't have database access, ensure listing tables in it fails
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    print("Listing datasets in: " + db)
                    print(conn.list_dataset_names(db))
                self.assertTrue('does not have privilege' in str(ex_ctx.exception))

                datasets = conn.list_datasets(db)
                self.assertEqual(len(datasets), 0, msg=str(datasets))

                for ddl in ['describe database %s', 'show tables in %s']:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        print(conn.execute_ddl(ddl % db))
                    self.assertTrue('does not have privilege' in str(ex_ctx.exception))

                if not skip_metadata_check:
                    dbs = conn.list_databases()
                    self.assertFalse(TEST_DB in dbs, msg=str(dbs))
        else:
            dbs = conn.list_databases()
            self.assertTrue(db in dbs, msg=str(dbs))
            names = conn.list_dataset_names(db)
            self.assertTrue('%s.%s' %(db, tbl) in names, msg=str(names))
            datasets = conn.list_datasets(db, name=tbl)
            self.assertEqual(len(datasets), 1)
            cols = self.__visible_cols(conn.list_datasets(db, name=tbl)[0].schema.cols)
            self.assertEqual(len(cols), num_cols)

            datasets = conn.list_datasets(db)
            self.assertTrue(len(datasets) >= 1, msg=str(datasets))

    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(AbacTest, cls).setUpClass()
        ctx = context()
        with ctx.connect() as conn:
            conn.delete_attribute('abac_test', 'v1')
            conn.create_attribute('abac_test', 'v1')
            conn.delete_attribute('abac_test', 'v2')
            conn.create_attribute('abac_test', 'v2')
            conn.delete_attribute('abac_test', 'v3')
            conn.create_attribute('abac_test', 'v3')
            conn.delete_attribute('abac_test', 'pii')
            conn.create_attribute('abac_test', 'pii')
            conn.delete_attribute('abac_test', 'email')
            conn.create_attribute('abac_test', 'email')
            conn.delete_attribute('abac_test', 'sales')
            conn.create_attribute('abac_test', 'sales')
            conn.delete_attribute('abac_test', 'marketing')
            conn.create_attribute('abac_test', 'marketing')

    @staticmethod
    def __cleanup(conn, dbs=None):
        """ Cleanups all the test state used in this test to "reset" the catalog.
            dbs can be specified to do the initialize over multiple databases.
            This can be used for tests that use multiple dbs (but makes the test
            take longer). By default, only load TEST_DB.
        """
        conn.execute_ddl("DROP ROLE IF EXISTS %s" % TEST_ROLE)
        conn.execute_ddl("CREATE ROLE %s" % TEST_ROLE)
        conn.execute_ddl("GRANT ROLE %s to GROUP %s" % (TEST_ROLE, TEST_USER))

        # Always drop both to prevent state from leaking
        conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % TEST_DB)
        conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % TEST_DB2)

        if not dbs:
            dbs = [TEST_DB]

        for db in dbs:
            conn.execute_ddl("CREATE DATABASE %s" % db)
            conn.execute_ddl("CREATE TABLE %s.%s(col1 int, col2 int, col3 int)" % \
                (db, TEST_TBL))
            conn.execute_ddl("CREATE TABLE %s.%s(col1 int, col2 int, col3 int)" % \
                (db, TEST_TBL2))
            conn.execute_ddl("CREATE VIEW %s.%s AS SELECT * FROM %s.%s" % \
                (db, TEST_VIEW, TEST_DB, TEST_TBL))

    @staticmethod
    def __revoke(conn, sql):
        """ Revokes a grant, transforming a GRANT statement to its corresponding REVOKE"""
        sql = sql.lower()
        sql = sql.replace('grant', 'revoke')
        sql = sql.replace('to role', 'from role')
        conn.execute_ddl(sql)

    @staticmethod
    def __fixture1(conn):
        """ Creates a representative test fixture. This will:
              Tag TEST_DB with department sales
              Tag TEST_DB2 with department marketing.
              Tag TEST_DB2.TEST_TBL with department sales.
              Tag some columns with pii tags, and some with another tag and a third
              column with no tags.
        """
        conn.assign_attribute('abac_test', 'sales', TEST_DB)
        conn.assign_attribute('abac_test', 'marketing', TEST_DB2)
        conn.assign_attribute('abac_test', 'sales', TEST_DB2, TEST_TBL)

        for db in [TEST_DB, TEST_DB2]:
            conn.assign_attribute('abac_test', 'pii', db, TEST_TBL, 'col1')
            conn.assign_attribute('abac_test', 'pii', db, TEST_TBL2, 'col1')
            conn.assign_attribute('abac_test', 'pii', db, TEST_VIEW, 'col1')
            conn.assign_attribute('abac_test', 'v1', db, TEST_TBL, 'col2')
            conn.assign_attribute('abac_test', 'v1', db, TEST_TBL2, 'col2')
            conn.assign_attribute('abac_test', 'v1', db, TEST_VIEW, 'col2')

    # Tests that grants are case insensitive for catalog objects, attributes and exprs.
    def test_case_insensitive_exprs(self):
        ctx = context()
        with ctx.connect() as conn:
            self.__cleanup(conn)
            # Grant with all lower case
            conn.execute_ddl(
                'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
                (TEST_DB, 'abac_test', 'v1', TEST_ROLE))

            # Try some duplicates, they should all fail
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                    (TEST_DB, 'abac_test.v1', TEST_ROLE))
            self.assertTrue('Grant already exists' in str(ex_ctx.exception))
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                    (TEST_DB, 'abac_TEST.v1', TEST_ROLE))
            self.assertTrue('Grant already exists' in str(ex_ctx.exception))
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                    (TEST_DB.upper(), 'not abac_test.v1', TEST_ROLE))
            self.assertTrue('Another grant with' in str(ex_ctx.exception))
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
                    (TEST_DB.upper(), 'Abac_test', 'v1', TEST_ROLE))
            self.assertTrue('Grant already exists' in str(ex_ctx.exception))
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                    (TEST_DB.upper(), 'NOT Abac_test.V1', TEST_ROLE))
            self.assertTrue('Another grant with' in str(ex_ctx.exception))
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
                    (TEST_DB.upper(), 'Abac_test', 'V1', TEST_ROLE))
            self.assertTrue('Grant already exists' in str(ex_ctx.exception))

    def test_in_print(self):
        ctx = context()
        with ctx.connect() as conn:
            self.__cleanup(conn)
            # Grant with all lower case
            conn.execute_ddl(
                'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE IN(%s.%s) TO ROLE %s' %\
                (TEST_DB, 'abac_test', 'v1', TEST_ROLE))
            conn.execute_ddl(
                'GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE NOT IN(%s.%s) TO ROLE %s' %\
                (TEST_DB, 'abac_test', 'v2', TEST_ROLE))

            result = conn.execute_ddl('SHOW GRANT ROLE %s' % TEST_ROLE)
            self.assertEqual(2, len(result))
            for r in result:
                expr = r[6]
                self.assertFalse('null' in expr)

    def test_db(self):
        ctx = context()
        with ctx.connect() as conn:
            self.__cleanup(conn)

            # Admin should see it but test user should not
            self.assertTrue(TEST_DB in conn.list_databases())
            ctx.enable_token_auth(token_str=TEST_USER)
            self.assertTrue('okera_sample' in conn.list_databases())
            self.assertFalse(TEST_DB in conn.list_databases())

            # Switch back to admin and assign abac_test.v1 to this db
            ctx.disable_auth()
            self.__grant_abac_db(conn, TEST_DB, 'abac_test', 'v1')

            # Test user should not have access to this DB but currently does.
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 0)

            # Assign the attribute
            ctx.disable_auth()
            conn.assign_attribute('abac_test', 'v1', TEST_DB)

            # List databases again, test user should have access now
            self.assertTrue(TEST_DB in conn.list_databases())
            ctx.enable_token_auth(token_str=TEST_USER)
            self.assertTrue('okera_sample' in conn.list_databases())
            self.assertTrue(TEST_DB in conn.list_databases())

            # Should have full access to both tables
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 3)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)
            self.__verify_tbl_access(conn, TEST_DB, TEST_VIEW, 3)

    def test_tbl_view(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                # Admin should see it but test user should not
                self.assertTrue(
                    '%s.%s' %(TEST_DB, ds) in conn.list_dataset_names(TEST_DB))
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

                # Switch back to admin and assign abac_test.v1 to this dataset
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'v1', TEST_DB, ds)
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')
                # List tables again, test user should have access now
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)

                # Unassign the attribute, testuser should not see it anymore
                ctx.disable_auth()
                conn.unassign_attribute('abac_test', 'v1', TEST_DB, ds)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

                # Assign the attribute again but revoke the grant, should not have access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'v1', TEST_DB, ds)
                self.__revoke_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

    def test_col(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                # Verify no access to start
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_VIEW, 0)

                # Switch back to admin and assign abac_test.v1 to just one column and
                # issue the grant.
                ctx.disable_auth()
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')
                conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col1')
                self.assertTrue(
                    '%s.%s' %(TEST_DB, ds) in conn.list_dataset_names(TEST_DB))

                # Verify user can just see one column on this one table
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 1)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0, has_db_access=True)

                # Assign attribute to other column
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col2')

                # Now should be able to see both.
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 2)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0, has_db_access=True)

                # Unassign one attribute
                ctx.disable_auth()
                conn.unassign_attribute('abac_test', 'v1', TEST_DB, ds, 'col1')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 1)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0, has_db_access=True)

                # Unassign other
                ctx.disable_auth()
                conn.unassign_attribute('abac_test', 'v1', TEST_DB, ds, 'col2')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0)

                # Revoke the grant
                ctx.disable_auth()
                self.__revoke_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0)

    def test_attr(self):
        ctx = context()
        with ctx.connect() as conn:
            for expr in ['abac_test.v1', 'in(abac_test.v1)',
                         'in(abac_test.v2, abac_test.v1)']:
                for ds in [TEST_TBL, TEST_VIEW]:
                    ctx.disable_auth()
                    self.__cleanup(conn)

                    # No tags, user can none
                    conn.execute_ddl(
                        'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE %s TO ROLE %s' %\
                        (TEST_DB, ds, expr, TEST_ROLE))
                    ctx.disable_auth()
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 0)

                    # Assign attribute to one column, user can see 1
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col1')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 1)

                    # Assign attribute to second column, user can see 2
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col2')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 2)

                    # Assign attribute to last column, user can see all 3
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col3')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 3)

    def test_not(self):
        ctx = context()
        with ctx.connect() as conn:
            for expr in ['not abac_test.v1', 'not in(abac_test.v1)',
                         'not in(abac_test.v2, abac_test.v1)',
                         'not not not abac_test.v1']:
                for ds in [TEST_TBL, TEST_VIEW]:
                    ctx.disable_auth()
                    self.__cleanup(conn)

                    # No tags, user can see all 3
                    conn.execute_ddl(
                        'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE %s TO ROLE %s' %\
                        (TEST_DB, ds, expr, TEST_ROLE))
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 3)

                    # Assign attribute to one column, user can see 2
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col1')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 2)

                    # Assign attribute to second column, user can see 1
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col2')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 1)

                    # Assign attribute to last column, user cannot see
                    ctx.disable_auth()
                    conn.assign_attribute('abac_test', 'v1', TEST_DB, ds, 'col3')
                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 0)

    # Verify abac privileges for role are removed when role is dropped
    def test_drop_role(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                # grant attribute privileges to role
                conn.assign_attribute('abac_test', 'v1', TEST_DB)
                self.__grant_abac_db(conn, TEST_DB, 'abac_test', 'v1')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')

                # Ensure 2 privileges
                result = self.__show_grant_abac_role(conn)
                self.assertEqual(len(result), 2)

                conn.execute_ddl("DROP ROLE %s" % TEST_ROLE)
                conn.execute_ddl("CREATE ROLE %s" % TEST_ROLE)

                # Ensure 0 privileges
                result = self.__show_grant_abac_role(conn)
                self.assertEqual(len(result), 0)
                conn.execute_ddl("DROP ROLE %s" % TEST_ROLE)

    def test_show(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)
                admin_before_count = self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON CATALOG')
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON DATABASE %s' % TEST_DB))
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON TABLE %s.%s' % \
                        (TEST_DB, ds)))
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON CATALOG' % TEST_ROLE))

                # Issue some grants, make sure they show correctly
                self.__grant_abac_db(conn, TEST_DB, 'abac_test', 'v1')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')

                self.assertEqual(admin_before_count + 2, self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON CATALOG'))
                self.assertEqual(2, self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON DATABASE %s' % TEST_DB))
                self.assertEqual(1, self.__ddl_count(
                    conn, 'SHOW GRANT ATTRIBUTE abac_test.v1 ON TABLE %s.%s' % \
                    (TEST_DB, ds)))
                self.assertEqual(
                    self.__ddl_count(conn, 'SHOW GRANT ROLE %s' % TEST_ROLE), 2)
                self.assertEqual(2, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON CATALOG' % TEST_ROLE))
                self.assertEqual(2, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON DATABASE %s' % (TEST_ROLE, TEST_DB)))
                self.assertEqual(1, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON TABLE %s.%s' % \
                    (TEST_ROLE, TEST_DB, ds)))
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON DATABASE %s' % \
                    (TEST_ROLE, 'okera_sample')))
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON TABLE %s.%s' % \
                    (TEST_ROLE, 'okera_sample', 'users')))

    def test_show_multiple_grants(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL]:
                ctx.disable_auth()
                self.__cleanup(conn)

                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v2')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v3')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'sales')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'marketing')

                self.assertEqual(5, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON TABLE %s.%s' % \
                    (TEST_ROLE, TEST_DB, ds)))

    def test_literal_expression(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                # Attribute expression without '='
                conn.execute_ddl(
                    'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s.%s) TO ROLE %s' %\
                    (TEST_DB, ds, 'abac_test', 'v1', TEST_ROLE))
                self.assertEqual(1, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON TABLE %s.%s' % \
                    (TEST_ROLE, TEST_DB, ds)))
                conn.assign_attribute('abac_test', 'v1', TEST_DB, ds)

                # List tables as test user
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0, has_db_access=True)

                # Not expression
                ctx.disable_auth()
                conn.execute_ddl(
                    'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(!%s.%s) TO ROLE %s' %\
                    (TEST_DB, TEST_TBL2, 'abac_test', 'v2', TEST_ROLE))
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Revoke one of them
                ctx.disable_auth()
                conn.execute_ddl(
                    'REVOKE SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s.%s) FROM ROLE %s' %\
                    (TEST_DB, ds, 'abac_test', 'v1', TEST_ROLE))
                self.assertEqual(0, self.__ddl_count(
                    conn, 'SHOW GRANT ROLE %s ON TABLE %s.%s' % \
                    (TEST_ROLE, TEST_DB, ds)))
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0, has_db_access=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Revoke other one
                ctx.disable_auth()
                conn.execute_ddl(
                    ('REVOKE SELECT ON TABLE %s.%s HAVING ATTRIBUTE(!%s.%s) ' +\
                    'FROM ROLE %s') %\
                    (TEST_DB, TEST_TBL2, 'abac_test', 'v2', TEST_ROLE))
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0)

    def test_table_column_and_not_and(self):
        ctx = context()
        with ctx.connect() as conn:
            self.__cleanup(conn)

            # Grant with and AND of ORs
            conn.execute_ddl(
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                (TEST_DB, TEST_TBL,
                 'abac_test.v1=true and (abac_test.v2 != true and abac_test.v3 != true)',
                 TEST_ROLE))

            # Assign v1, should see both columns
            conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL)
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 3)

            # Tag one of the columns with v2, it should be hidden now
            ctx.disable_auth()
            conn.assign_attribute('abac_test', 'v2', TEST_DB, TEST_TBL, 'col1')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 2)

            # Tag the other column with the other tag, both should be hidden
            ctx.disable_auth()
            conn.assign_attribute('abac_test', 'v3', TEST_DB, TEST_TBL, 'col2')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)

    def test_table_column_and_or(self):
        ctx = context()
        with ctx.connect() as conn:
            self.__cleanup(conn)

            # Assign abac_test.v1 to this table
            conn.execute_ddl(
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                (TEST_DB, TEST_TBL,
                 'abac_test.v1 and IN (abac_test.v2, abac_test.v3)',
                 TEST_ROLE))

            # Should see nothing
            conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL)
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 0)

            # Tag one of the columns with v2, it should be hidden now
            ctx.disable_auth()
            conn.assign_attribute('abac_test', 'v2', TEST_DB, TEST_TBL, 'col1')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)

            # Tag the other column with the other tag, both should be visible
            ctx.disable_auth()
            conn.assign_attribute('abac_test', 'v3', TEST_DB, TEST_TBL, 'col2')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 2)

            # Unassign one of the columns with v2, it should be hidden now
            ctx.disable_auth()
            conn.unassign_attribute('abac_test', 'v2', TEST_DB, TEST_TBL, 'col1')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)

            # Unassign other columns with v1, it should be hidden now
            ctx.disable_auth()
            conn.unassign_attribute('abac_test', 'v3', TEST_DB, TEST_TBL, 'col2')
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 0)

    def test_das_3117(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                # grant != pii at the view level
                conn.execute_ddl(
                    'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE(%s) TO ROLE %s' %\
                    (TEST_DB, ds, 'not abac_test.pii', TEST_ROLE))
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)

                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

    def test_das_3119(self):
        ctx = context()
        with ctx.connect() as conn:
            ctx.disable_auth()
            self.__cleanup(conn)

            # Assign the attribute to a col on the table and view
            conn.assign_attribute('abac_test', 'pii', TEST_DB, TEST_TBL, 'col1')
            conn.assign_attribute('abac_test', 'pii', TEST_DB, TEST_VIEW, 'col1')

            conn.execute_ddl(
                'GRANT SELECT ON TABLE %s.%s HAS ATTRIBUTE(%s) TO ROLE %s' %\
                (TEST_DB, TEST_TBL, 'abac_test.pii', TEST_ROLE))
            conn.execute_ddl(
                'GRANT SELECT ON TABLE %s.%s HAS ATTRIBUTE(%s) TO ROLE %s' %\
                (TEST_DB, TEST_VIEW, 'abac_test.pii', TEST_ROLE))

            # User should have access to this column on both
            ctx.enable_token_auth(token_str=TEST_USER)

            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)
            self.__verify_tbl_access(conn, TEST_DB, TEST_VIEW, 1)

    def test_grant1(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE ' +\
                    'abac_test.SALES AND NOT abac_test.PII TO ROLE %s') %\
                    (TEST_DB, ds, TEST_ROLE))

                # Should not have any access
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

                # Just assign sales
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0, has_db_access=True)

                # Also assign PII, now should have no access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)

    def test_grant2(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                    'IN (abac_test.SALES, abac_test.MARKETING) ' +\
                    'AND NOT IN (abac_test.pii) TO ROLE %s') %\
                    (TEST_DB, TEST_ROLE))

                # Should not have any access
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0)

                # Assign sales to one and marketing to another, should have both
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'marketing', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Assign PII to one, now should not have access to that one
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0, has_db_access=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Assign PII to the other, now should have no access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0, skip_metadata_check=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0,
                                         skip_metadata_check=True)

    def test_grant3(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                    'IN (abac_test.SALES, abac_test.MARKETING) AND ' +\
                    'NOT abac_test.PII AND NOT abac_test.v1 TO ROLE %s') %\
                    (TEST_DB, TEST_ROLE))

                # Assign sales to one and marketing to another, should have both
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'marketing', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Assign PII to one and v1 to the other, should not have access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0, skip_metadata_check=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0,
                                         skip_metadata_check=True)

    def test_grant4(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                    'IN (abac_test.SALES, abac_test.MARKETING) ' +\
                    'AND NOT IN(abac_test.PII, abac_test.v1) TO ROLE %s') % \
                    (TEST_DB, TEST_ROLE))

                # Assign sales to one and marketing to another, should have both
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'marketing', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Assign PII to one and v1 to the other, should not have access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 0,
                                         skip_metadata_check=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0,
                                         skip_metadata_check=True)

    def test_grant5(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                    'IN (abac_test.SALES, abac_test.MARKETING) ' +\
                    'AND NOT IN(abac_test.PII, abac_test.v1) TO ROLE %s') % \
                    (TEST_ROLE))

                # Assign sales to one and marketing to another, should have both
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'marketing', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 3)

                # Assign PII to one and v1 to the other, should not have access
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds)
                conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL2)
                ctx.enable_token_auth(token_str=TEST_USER)

                self.__verify_tbl_access(conn, TEST_DB, ds, 0,
                                         skip_metadata_check=True)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 0,
                                         skip_metadata_check=True)

    def test_grant7(self):
        grants = []
        grants.append(
            ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
            '(NOT abac_test.PII OR NOT abac_test.email) TO ROLE %s') % \
            (TEST_ROLE))
        grants.append(
            ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
            'NOT (abac_test.PII AND abac_test.email) TO ROLE %s') % \
            (TEST_ROLE))

        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                for sql in grants:
                    ctx.disable_auth()
                    self.__cleanup(conn)

                    conn.execute_ddl(sql)

                    # Assign sales to one and marketing to another, should have both
                    conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                    conn.assign_attribute('abac_test', 'pii', TEST_DB, ds, 'col1')
                    conn.assign_attribute('abac_test', 'pii', TEST_DB, ds, 'col2')
                    conn.assign_attribute('abac_test', 'email', TEST_DB, ds, 'col2')
                    conn.assign_attribute('abac_test', 'email', TEST_DB, ds, 'col3')

                    ctx.enable_token_auth(token_str=TEST_USER)
                    self.__verify_tbl_access(conn, TEST_DB, ds, 2)

    def test_grant6(self):
        # Regression test for DAS-2928
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL]:
                ctx.disable_auth()
                self.__cleanup(conn)

                conn.execute_ddl(
                    ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                    'IN (abac_test.SALES, abac_test.MARKETING) ' +\
                    'AND NOT IN(abac_test.PII, abac_test.EMAIL) TO ROLE %s') % \
                    (TEST_ROLE))

                # Assign sales to the table
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'sales', TEST_DB, ds)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 3)

                # Assign PII to one and email to two different columns. These columns
                # should be hidden but the third column should be visible
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'pii', TEST_DB, ds, 'col1')
                conn.assign_attribute('abac_test', 'email', TEST_DB, ds, 'col2')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, ds, 1)

    def test_das_3122(self):
        # grant attribute privileges for database to be removed
        ctx = context()
        with ctx.connect() as conn:
            ctx.disable_auth()
            self.__cleanup(conn)

            conn.assign_attribute('abac_test', 'v1', TEST_DB)
            self.__grant_abac_db(conn, TEST_DB, 'abac_test', 'v1')
            result = self.__show_grant_abac_role(conn)
            self.assertEqual(len(result), 1)

            # Drop and Ensure privilege is removed
            conn.execute_ddl("DROP DATABASE %s CASCADE INCLUDING PERMISSIONS " %\
                             TEST_DB)
            result = self.__show_grant_abac_role(conn)
            self.assertEqual(len(result), 0)

    def test_das_3122_tbl(self):
        ctx = context()
        with ctx.connect() as conn:
            for ds in [TEST_TBL, TEST_VIEW]:
                ctx.disable_auth()
                self.__cleanup(conn)

                TV = 'TABLE'
                if ds == TEST_VIEW:
                    TV = 'VIEW'

                # grant attribute privileges for table to be removed
                conn.assign_attribute('abac_test', 'v1', TEST_DB)
                self.__grant_abac_db(conn, TEST_DB, 'abac_test', 'v1')
                self.__grant_abac_tbl(conn, TEST_DB, ds, 'abac_test', 'v1')

                # Ensure privileges for db and table
                result = self.__show_grant_abac_role(conn)
                self.assertEqual(len(result), 2)
                self.assertTrue("database" in result[0][0])
                self.assertTrue("table" in result[1][0])

                conn.execute_ddl("DROP %s %s.%s INCLUDING PERMISSIONS" %\
                                 (TV, TEST_DB, ds))

                # Ensure privilege for DB only. Table privilege should have been dropped.
                result = self.__show_grant_abac_role(conn)
                self.assertEqual(len(result), 1)
                self.assertTrue("database" in result[0][0])

    def test_fixture1(self):
        ctx = context()
        with ctx.connect() as conn:
            ctx.disable_auth()
            self.__cleanup(conn, [TEST_DB, TEST_DB2])
            self.__fixture1(conn)

            # Grants that should all result in the same policies for these tests.
            grants = []
            grants.append(
                [('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                (TEST_ROLE)])
            grants.append(
                [('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES=true) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                (TEST_ROLE)])
            grants.append(
                [('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES!=false) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                (TEST_ROLE)])
            grants.append(
                [('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                (TEST_DB, TEST_ROLE),
                 ('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                 (TEST_DB2, TEST_ROLE)])
            grants.append(
                [('GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                (TEST_DB, TEST_ROLE),
                 ('GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE ' +\
                'IN (abac_test.SALES) AND NOT IN(abac_test.PII) TO ROLE %s') % \
                 (TEST_DB2, TEST_TBL, TEST_ROLE)])

            for grant in grants:
                for sql in grant:
                    conn.execute_ddl(sql)
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 2)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 2)
                self.__verify_tbl_access(conn, TEST_DB, TEST_VIEW, 2)
                self.__verify_tbl_access(conn, TEST_DB2, TEST_TBL, 2)
                self.__verify_tbl_access(conn, TEST_DB2, TEST_TBL2, 0, has_db_access=True)
                self.__verify_tbl_access(conn, TEST_DB2, TEST_VIEW, 0, has_db_access=True)

                ctx.disable_auth()
                for sql in grant:
                    self.__revoke(conn, sql)

            # Grant access to (sales and pii) or (marketing and v1)
            conn.execute_ddl(
                ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
                'abac_test.SALES and abac_test.pii OR ' +\
                'abac_test.MARKETING and abac_test.v1 ' +\
                'TO ROLE %s') % (TEST_ROLE))
            ctx.enable_token_auth(token_str=TEST_USER)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)
            self.__verify_tbl_access(conn, TEST_DB, TEST_TBL2, 1)
            self.__verify_tbl_access(conn, TEST_DB, TEST_VIEW, 1)
            self.__verify_tbl_access(conn, TEST_DB2, TEST_TBL, 2)
            self.__verify_tbl_access(conn, TEST_DB2, TEST_TBL2, 1)
            self.__verify_tbl_access(conn, TEST_DB2, TEST_VIEW, 1)

    def test_white_list(self):
        # Test with a pii tag on the table and then white list on columns
        ctx = context()

        grants = []
        grants.append(
            ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
            'abac_test.SALES and ' + \
            '(not abac_test.pii OR abac_test.v1 or abac_test.v2) ' +\
            'TO ROLE %s') % (TEST_ROLE))
        grants.append(
            ('GRANT SELECT ON CATALOG HAVING ATTRIBUTE ' +\
            'abac_test.SALES and ' + \
            'not (abac_test.pii and not abac_test.v1 and not abac_test.v2) ' +\
            'TO ROLE %s') % (TEST_ROLE))

        with ctx.connect() as conn:
            for sql in grants:
                ctx.disable_auth()
                self.__cleanup(conn, [TEST_DB, TEST_DB2])

                # Creates a representative test fixture. This will:
                # Tag TEST_DB with department sales
                # Tag TEST_DB.TEST_TBL with pii tags.
                conn.assign_attribute('abac_test', 'sales', TEST_DB)
                conn.assign_attribute('abac_test', 'pii', TEST_DB, TEST_TBL)

                # Issue the grant
                conn.execute_ddl(sql)

                ctx.enable_token_auth(token_str=TEST_USER)
                #self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 0, has_db_access=True)

                # Tag some of the columns with non-pii tags.
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL, 'col1')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 1)

                # Tag some of the columns with non-pii tags.
                ctx.disable_auth()
                conn.assign_attribute('abac_test', 'v2', TEST_DB, TEST_TBL, 'col2')
                conn.assign_attribute('abac_test', 'v1', TEST_DB, TEST_TBL, 'col3')
                conn.assign_attribute('abac_test', 'v2', TEST_DB, TEST_TBL, 'col3')
                ctx.enable_token_auth(token_str=TEST_USER)
                self.__verify_tbl_access(conn, TEST_DB, TEST_TBL, 3)

    def test_get_access_permissions(self):
        ctx = context()

        with ctx.connect() as conn:
            ctx.disable_auth()
            self.__cleanup(conn)

            conn.execute_ddl((
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE abac_test.v1 ' +
                'TO ROLE %s;') % (TEST_DB, TEST_TBL, TEST_ROLE))
            conn.execute_ddl((
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE ' +
                'abac_test.v2 TO ROLE %s;') % (TEST_DB, TEST_TBL2, TEST_ROLE))
            conn.execute_ddl((
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE ' +
                'abac_test.v3 TO ROLE %s;') % (TEST_DB, TEST_TBL2, TEST_ROLE))

            # Test that TEST_DB.TEST_TBL has expected attribute expressions
            params = _thrift_api.TGetAccessPermissionsParams()
            params.users_or_groups = []
            params.database = TEST_DB
            params.filter = TEST_TBL
            #pylint: disable=protected-access
            permissions = \
                conn._underlying_client().GetAccessPermissions(params).permissions
            #pylint: enable=protected-access
            test_role_permissions = [
                perm for perm in permissions
                if perm.database == TEST_DB and perm.table == TEST_TBL]
            self.assertEqual(len(test_role_permissions), 1)
            self.assertEqual(len(test_role_permissions[0].attribute_expression), 1)
            self.assertEqual(
                test_role_permissions[0].attribute_expression[0].__dict__,
                {
                    'role_name' : TEST_ROLE,
                    'expression' : 'abac_test.v1'})

            # Test that TEST_DB.TEST_TBL2 has expected attribute expressions
            params = _thrift_api.TGetAccessPermissionsParams()
            params.users_or_groups = []
            params.database = TEST_DB
            params.filter = TEST_TBL2
            #pylint: disable=protected-access
            permissions = \
                conn._underlying_client().GetAccessPermissions(params).permissions
            #pylint: enable=protected-access
            test_role_permissions = [
                perm for perm in permissions
                if perm.database == TEST_DB and perm.table == TEST_TBL2]
            self.assertEqual(len(test_role_permissions), 2)
            self.assertEqual(len(test_role_permissions[0].attribute_expression), 1)
            self.assertEqual(len(test_role_permissions[1].attribute_expression), 1)

            roles = [
                exp.role_name
                for exp in test_role_permissions[0].attribute_expression]
            roles = roles + [
                exp.role_name
                for exp in test_role_permissions[1].attribute_expression]
            expressions = [
                exp.expression for exp in test_role_permissions[0].attribute_expression]
            expressions = expressions + [
                exp.expression for exp in test_role_permissions[1].attribute_expression]

            self.assertEqual(set(roles), set([TEST_ROLE]))
            self.assertEqual(set(expressions), set(['abac_test.v2', 'abac_test.v3']))

    def test_drop_db(self):
        ctx = context()
        with ctx.connect() as conn:
            ctx.disable_auth()

            # Create our desired state of a single DB and table in it and a unique
            # role and user
            conn.execute_ddl("DROP TABLE IF EXISTS %s.%s" % (TEST_DB_DROP, TEST_TBL_DROP))
            conn.execute_ddl("DROP DATABASE IF EXISTS %s" % (TEST_DB_DROP))
            conn.execute_ddl("CREATE DATABASE %s" % (TEST_DB_DROP))
            conn.execute_ddl("CREATE TABLE %s.%s (i int)" % (TEST_DB_DROP, TEST_TBL_DROP))
            conn.execute_ddl("DROP ROLE IF EXISTS %s" % (TEST_ROLE_DROP))
            conn.execute_ddl("CREATE ROLE %s" % (TEST_ROLE_DROP))
            conn.execute_ddl("GRANT ROLE %s to GROUP %s"
                             % (TEST_ROLE_DROP, TEST_USER_DROP))

            conn.execute_ddl((
                'GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE NOT IN (abac_test.v1) ' +
                'TO ROLE %s;') % (TEST_DB_DROP, TEST_TBL_DROP, TEST_ROLE_DROP))

            # Assign the tag to the table
            conn.assign_attribute('abac_test', 'pii', TEST_DB_DROP, TEST_TBL_DROP)

            # Ensure 'show databases' works for admin
            dbs = [db[0] for db in conn.execute_ddl('show databases')]
            self.assertTrue(TEST_DB_DROP in dbs)

            ctx.enable_token_auth(token_str=TEST_USER_DROP)

            # Ensure 'show databases' works for test user
            dbs = [db[0] for db in conn.execute_ddl('show databases')]
            self.assertTrue(TEST_DB_DROP in dbs)

            ctx.disable_auth()

            # Drop the DB and table
            conn.execute_ddl("DROP TABLE IF EXISTS %s.%s" % (TEST_DB_DROP, TEST_TBL_DROP))
            conn.execute_ddl("DROP DATABASE IF EXISTS %s" % (TEST_DB_DROP))

            ctx.enable_token_auth(token_str=TEST_USER_DROP)

            # Ensure 'show databases' works for test user (db shouldn't be there)
            dbs = [db[0] for db in conn.execute_ddl('show databases')]
            self.assertTrue(TEST_DB_DROP not in dbs)



if __name__ == "__main__":
    unittest.main()
